// waifusearch, the IB searcher
// ============================
// -This software is designed to search an imageboard's text.

// -Filename: main.cpp
// -The program's entry point, and implementation details.

#include <algorithm>
#include <chrono>
#include <cstdint>
#include <filesystem>
#include <fstream>
#include <functional>
#include <iomanip>
#include <iostream>
#include <iterator>
#include <map>
#include <numeric>
#include <set>
#include <sstream>
#include <string>
#include <utility>
#include <vector>

#if __has_include(<execution>)  // RPi will only run single-threaded for now
#include <execution>
using std::execution::par_unseq;
#endif

#include "Timer.hpp"
#include "rw_cli_args.hpp"
#include "rw_text_utils.hpp"
#include "user_flags.hpp"

// alias'
namespace fs    = std::filesystem;
using file_path = fs::path;
using dir_iter  = fs::directory_iterator;
using Ms        = std::chrono::milliseconds;
using Us        = std::chrono::microseconds;
//
using Pair_ints     = std::pair<uint32_t, uint32_t>;
using Set_pair_ints = std::set<Pair_ints>;
using Vec_pair_ints = std::vector<Pair_ints>;
//
using Map_words_2_posts = std::map<std::string, Set_pair_ints>;
using Vec_words_2_posts = std::vector<std::pair<std::string, Vec_pair_ints>>;
using Results_w_offset  = std::vector<std::pair<Pair_ints, long>>;

// -system globals:
std::map<uint32_t, string> thrd_subjs{};
std::map<uint32_t, string> all_msgs{};
std::string                search{};
std::vector<std::string>   unfounds{};
bool                       have_results{};

// usings
using std::accumulate;
using std::back_inserter;
using std::cbegin;
using std::cend;
using std::cerr;
using std::cin;
using std::cout;
using std::for_each;
using std::getline;
using std::left;
using std::make_pair;
using std::map;
using std::ofstream;
using std::out_of_range;
using std::pair;
using std::set;
using std::set_intersection;
using std::setw;
using std::stable_sort;
using std::string;
using std::stringstream;
using std::to_string;
using std::vector;
using std::chrono::duration_cast;
//
using rw::Timer;

/**-----------------------------------------------------------------------------
 @brief    Clean any unfound terms from the global 'search' string, if present.
*/
void clean_unfounds() {
  //
  if (! unfounds.empty()) {
    //
    for (auto& term : unfounds) {
      rw::squash_n_trim(term, ' ');
      rw::rm_substrings(search, term);
    }

    rw::squash_n_trim(search, ' ');
    unfounds.clear();
  }
}

/**-----------------------------------------------------------------------------
 @brief    Display the iterated, formatted, output of a search bin's results to
 console, noting the character offset for an ordered (IE, exact match) result,
 if present. The display includes the assembled URI link (whether crosslink or
 hyperlink) for each post.

 @param    vtp_o [IN] The search bin to output.
*/
void disp_results(const Results_w_offset& vtp_o) {
  //
  const uint32_t pad_len{32};
  cout << rw::make_pad_str("\n   THREAD SUBJECT", pad_len)
       << "        POST LINK\n";

  uint32_t prev_t{0};

  for (const auto& tp_o : vtp_o) {
    const auto [tp, offset]{tp_o};
    const auto [t, p]{tp};

    string subj{};
    if (t != prev_t)  // display subject for each new thread in list
      subj = rw::make_pad_str(thrd_subjs.at(t), pad_len);
    else  // no need to repeat the same thread subject, just 'ditto' it
      subj = rw::make_pad_str("   \" ", pad_len);

    // DESIGN: use C++20 std::format when available

    string   uri{};
    uint32_t uri_pad_len{};

    if (make_hyperlinks) {
      uri = {"  -> https://alogs.theguntretort.com/" + board_nm + "/res/" +
             to_string(t) + ".html#" + to_string(p)};
      uri_pad_len = 68;

    } else {  // use board crosslinks instead
      uri         = "   >>" + to_string(p);
      uri_pad_len = 13;
    }

    // the actual output statement
    cout << subj
         << rw::make_pad_str(uri, uri_pad_len)
         // display the character offset of an ordered result, if present
         << ((offset >= 0) ? "@ch: ~" + to_string(offset + 1) : "") << '\n';

    prev_t = t;
  }
}

/**-----------------------------------------------------------------------------
 @brief    Front end to marshal formatted display of 'ordered' & 'unordered'
 search bins out to console, noting total results count.

 @param    ordered [IN] The ordered bin.
 @param    unordered [IN] The unordered bin.
 @param    vtp_sz [IN] The total results count.
*/
void disp_head_foot(const Results_w_offset& ordered,
                    const Results_w_offset& unordered,
                    const size_t            vtp_sz) {
  const auto ord_sz   = ordered.size();
  const auto unord_sz = unordered.size();

  const bool have_ord{ord_sz > 0};
  const bool have_unord{unord_sz > 0};

  if (have_ord) {
    if (have_unord) {  // only display ORDERED header if both types are present
      cout << "\n                  ORDERED:\n                  ========";
    }
    disp_results(ordered);
  }

  if (have_unord) {
    cout << "\n                  UN-ORDERED:\n                  "
            "===========";
    disp_results(unordered);
  }

  cout << "\n' " << search <<

      // only segregate counts if UNORDERED results are present
      (have_unord
           ? " '  [" + to_string(ord_sz) + " : " + to_string(unord_sz) + "] = "
           : " '  =  ")

       << vtp_sz << " results\n";
}

/**-----------------------------------------------------------------------------
 @brief    Search the intersection vector 'vtp' for exact, ordered matches with
 the user's search terms, from the text container holding all messages. Exact
 orderings are stored into one container bin, and inexact into another. They are
 [both] then further processed.

 @param    vtp [IN] The vector of intersected thread_id+post_id pairs to search.
*/
void search_ordered(const Vec_pair_ints& vtp) {
  Timer t{"ord"};

  Results_w_offset ordered{}, unordered{};

  for (const auto& tp : vtp) {
    const auto [t, p]{tp};
    const auto msg = all_msgs.at(p);  // read in this post's message text

    // search message for an exact, ordered match with user's search terms
    const auto it =
        std::search(cbegin(msg), cend(msg),
                    std::boyer_moore_searcher(cbegin(search), cend(search)));

    if (it != cend(msg)) {  // exact order match found
      const auto char_offset = it - cbegin(msg);
      ordered.emplace_back(make_pair(tp, char_offset));

    } else
      unordered.emplace_back(make_pair(tp, -1));
  }

  t.stop();

  disp_head_foot(ordered, unordered, vtp.size());
}

//----------------------------------
// DESIGN:
// -Later, extend this approach to start doing near-miss spellings
// (lev_dist?), and synonyms (for vague recollections, etc.)
//

/**-----------------------------------------------------------------------------
 @brief    Intersect all the thread_id+post_id pairs in the ascending-sorted
 vector 'vw_vtp_valids' to find the complete intersection of all posts each
 containing all words in the vector. This intersection vector is then further
 processed.

 @param    vw_vtp_valids [IN] The vector of thread_id+post_id pairs to intersect
*/
void find_intersects(const Vec_words_2_posts& vw_vtp_valids) {
  Timer t{"int"};

  // load the first (shortest) vec to 'prime-the-pump'
  auto [w, vtp] = vw_vtp_valids[0];

  const auto terms_cnt = vw_vtp_valids.size();

  if (terms_cnt > 1) {  // have multiple terms, therefore perform intersections
    Vec_pair_ints tmp_vtp{};

    // note: indexing starts at 1, 0th already loaded above
    for (size_t i{1}; i < terms_cnt; ++i) {
      const auto& [w, next_vtp] = vw_vtp_valids[i];

      set_intersection(
#if __cpp_lib_parallel_algorithm >= 201603
          par_unseq,
#endif
          cbegin(vtp), cend(vtp),            //
          cbegin(next_vtp), cend(next_vtp),  //
          back_inserter(tmp_vtp));

      vtp = tmp_vtp;  // copy current result for next intersection
      tmp_vtp.clear();
    }
  }
  // Note: the case of a single term was set during 'prime-the-pump' statement

  t.stop();

  search_ordered(vtp);
}

/**-----------------------------------------------------------------------------
 @brief    Display located terms 'vw_vtp_valids' counts, formatted, to console.

 @param    vw_vtp_valids [IN] The vector of terms + counts to display.
 @param    pad_len [IN] A number specifying the display string lengths.
*/
void disp_terms(const Vec_words_2_posts& vw_vtp_valids,
                const uint32_t           pad_len) {
  //
  for (const auto& valid : vw_vtp_valids) {
    const auto& [term, tp]{valid};
    cout << " term:  " << rw::make_pad_str(term, pad_len)
         << "  count:  " << tp.size() << '\n';
  }

  cout << "# terms found:  " << vw_vtp_valids.size() << '\n';
}

/**-----------------------------------------------------------------------------
 @brief    Sort the vector 'io_vw_vtp_counts' by post counts, ascending is
 default.

 @param    io_vw_vtp_counts [IN/OUT] The target vector to sort.
 @param    descending [IN] A flag determining sorting mode.
*/
void sort_counts(Vec_words_2_posts& io_vw_vtp_counts,
                 const bool         descending = false) {
  Timer t{"srt"};

  stable_sort(
#if __cpp_lib_parallel_algorithm >= 201603
      par_unseq,
#endif
      begin(io_vw_vtp_counts), end(io_vw_vtp_counts),
      [descending](const auto& a, const auto& b) {
        const auto& [wa, tp_a]{a};
        const auto& [wb, tp_b]{b};

        if (descending)
          return tp_a.size() > tp_b.size();
        else
          return tp_a.size() < tp_b.size();
      });
}

/**-----------------------------------------------------------------------------
 @brief    Copy the word@post map 'w_tp's contents into vector 'io_vw_vtp'.

 @param    w_tp [IN] The source map to copy.
 @param    io_vw_vtp [IN/OUT] The target vector to populate.
*/
void cp_map_2_vec(const Map_words_2_posts& w_tp, Vec_words_2_posts& io_vw_vtp) {
  Timer t{"cpy"};

  for_each(
#if __cpp_lib_parallel_algorithm >= 201603
      par_unseq,
#endif
      cbegin(w_tp), cend(w_tp), [&](const auto& e) {
        const auto& [w, stp]{e};
        io_vw_vtp.emplace_back(
            make_pair(w, Vec_pair_ints{cbegin(stp), cend(stp)}));
      });
}

/**-----------------------------------------------------------------------------
 @brief    Iterate the collection of unfound terms and display them to console.
*/
void disp_unfounds() {
  if (unfounds.size() > 0) {  // global
    cout << '\n';

    for (const auto& unfound : unfounds) {
      cout << " term:  " << unfound << "  ** not found\n";
    }
  }
}

/**-----------------------------------------------------------------------------
 @brief    Search the word@post map 'w_tp' for every matching term sourced from
 the vector of strings 'terms'. Any results are then copied into other
 containers (map->map->vector), sorted, and further processed.

 @param    w_tp [IN] The map to search.
 @param    terms [IN] The vector of strings to locate.
*/
void locate_terms(const Map_words_2_posts&        w_tp,
                  const std::vector<std::string>& terms) {
  Timer t{"loc"};

  const uint32_t    pad_len{27};
  Map_words_2_posts uniques{};

  // attempt a term locate inside the w_tp and if found, copy it's thrd+post set
  for_each(
#if __cpp_lib_parallel_algorithm >= 201603
      par_unseq,
#endif
      cbegin(terms), cend(terms), [&](const auto& term) {
        try {
          // ****
          // NOTE: This 1 statement constitutes most of the algorithm's workload
          uniques[term] =
              w_tp.at(term);  // attempt to locate & thence to copy out set data
          // ****
        } catch (const out_of_range& e) {
          unfounds.emplace_back(rw::make_pad_str(term, pad_len));  // global
        }
      });

  t.stop();

  // display all unfounds first thing
  if (! unfounds.empty()) {
    disp_unfounds();
    clean_unfounds();  // remove unfounds from search string too
  }

  if (! uniques.empty()) {  // did we find anything?
    Vec_words_2_posts vw_vtp_valids{};

    cp_map_2_vec(uniques, vw_vtp_valids);
    sort_counts(vw_vtp_valids);  // sort terms by post counts, low to high

    if (show_terms)
      disp_terms(vw_vtp_valids, pad_len);

    find_intersects(vw_vtp_valids);

    have_results = true;  // global
  } else {
    have_results = false;

    if (! unfounds.empty())  // just in case we have _all_ unfounds,
      unfounds.clear();      // clear container for next search round
  }
}

/**-----------------------------------------------------------------------------
 @brief    Perform an over-reaching character cleanup of the 'io_str' string to
 partially try dealing with all the unexpectedly random text not uncommon in IB
 post messages. In addition, all alphabetic chars are lower-cased here.

 @param    io_str [IN/OUT] The string to clean.
*/
void scrub_chars(std::string& io_str) {
  const vector<pair<char, char>> bad_goods{
      {'/', ' '}, {'.', ' '}, {'!', ' '},  {'?', ' '},  {'&', ' '},  {'^', ' '},
      {'$', ' '}, {'%', ' '}, {'#', ' '},  {'@', ' '},  {'*', ' '},  {'~', ' '},
      {'"', ' '}, {'=', ' '}, {'\0', ' '}, {'\n', ' '}, {'(', ' '},  {')', ' '},
      {'[', ' '}, {']', ' '}, {'<', ' '},  {'>', ' '},  {'{', ' '},  {'}', ' '},
      {',', ' '}, {':', ' '}, {'\\', ' '}, {';', ' '},  {'\t', ' '}, {'-', ' '},
      {'_', ' '}, {'`', ' '}, {'|', ' '}};

  for (auto [bad, good] : bad_goods)
    rw::repl_bad_char(io_str, bad, good);

  rw::to_lower(io_str);
}

// NOTES:

// DESIGN: Other than the legitimate single-letters 'a' & 'I', (and slang 'U'
// & 'Y', and the programming language 'C'), all the letters should be zapped if
// they are coming up as individual words, similar to the numbers (though
// numbers can at least actually have legit context. the letters thing is
// probably more of an artifact of our quick & dirty overfiltering approach
// here).

// DESIGN: Rework to create and use a 'rm_all_nums()' but guard it with a test
// first against a list of specific numbers like years, 9001, etc. to keep.

/**-----------------------------------------------------------------------------
 @brief    Clear the 'io_str' string if it matches unexpectedly random text not
 uncommon in IB post messages.

 @param    io_str [IN/OUT] The string potentially to clear.
*/
void scrub_string(std::string& io_str) {
  const vector<string> bad_strings{
      "www", "n",  "v", "'", "b",  "w", "o",   "h",    "t",    "s", "e",
      "x",   "fl", "+", "m", "fo", "r", "p",   "d",    "f",    "g", "y",
      "nn",  "u",  "–", "z", "q",  "j", "iii", "apos", "quot", "0", "1",
      "2",   "3",  "4", "5", "6",  "7", "8",   "9"};

  for (const auto& bad_string : bad_strings)
    if (io_str == bad_string)
      io_str.clear();
}

/**-----------------------------------------------------------------------------
 @brief    Wrapper to trim the 'io_str' string of any target chars (currently
 <spc> & <'>) .

 @param    io_str [IN/OUT] The string to clean.
*/
void trim_chars(std::string& io_str) {
  rw::trim_lr(io_str, ' ');
  rw::trim_lr(io_str, '\'');
}

/**-----------------------------------------------------------------------------
 @brief    Wrapper to remove unwanted chars from the 'io_str' string, and to
 consolidate consecutive chars (currently <spc> & <'>) down to just one per.

 @param    io_str [IN/OUT] The string to clean.
*/
void scrub_rm_dupes(std::string& io_str) {
  scrub_chars(io_str);
  rw::rm_dupes(io_str, ' ');
  rw::rm_dupes(io_str, '\'');
}

/**-----------------------------------------------------------------------------
 @brief    Parse the input string 'text' into a vector of cleaned strings.

 @param    text [IN] The string to parse.
 @returns  The vector of strings after parsing.
*/
std::vector<std::string> scrub_n_extract(const std::string& text) {
  string temp{text};
  scrub_rm_dupes(temp);

  stringstream   message{temp};
  vector<string> words{};
  //
  while (message >> temp) {  // parses on whitespace
    trim_chars(temp);
    scrub_string(temp);

    if (! temp.empty())
      words.push_back(temp);
  }

  return words;
}

/**-----------------------------------------------------------------------------
 @brief    Iterate the collection of chrono timings and display them to console,
 displayed as raw microseconds and by percentages.
*/
void disp_pct_timings() {
  const string divider(80, '-');
  cout << divider << '\n';

  // Raw timings display
  cout << rw::make_pad_str(" total:", 11);
  vector<long> us_timings{};
  //
  for (const auto& [time_pt, name] : rw::named_timings) {  // global
    const auto timing = duration_cast<Us>(time_pt.time_since_epoch()).count();

    cout << rw::make_pad_str(name, 3) << ':'
         << rw::make_pad_str(to_string(timing), 7);
    us_timings.push_back(timing);
  }
  cout << "us\n ";

  const auto us_sum = accumulate(cbegin(us_timings), cend(us_timings), 0);
  cout << rw::make_pad_str(to_string(us_sum) + " us", 13) << ':';

  // Percentages display
  size_t n{0};
  //
  for (const auto timing : us_timings) {
    cout << (++n < us_timings.size()
                 ? rw::make_pad_str(rw::make_pct_str(timing, us_sum), 10) + ':'
                 : rw::make_pad_str(rw::make_pct_str(timing, us_sum), 7));
  }
  cout << "%\n";

  cout << divider << '\n';
}

/**-----------------------------------------------------------------------------
 @brief    Front end to search-loop the word@post map 'w_tp' for user's query
 terms.

 @param    w_tp [IN] The word/thread+post map to search through.
*/
void do_search(Map_words_2_posts& w_tp) {
  string srch_phrase{};
  while (true) {
    cout << "\n>search:  ";
    getline(cin, srch_phrase);

    if (srch_phrase == "q")
      break;
    else {
      Timer t{"sea"};

      const auto words = scrub_n_extract(srch_phrase);
      search           = rw::assemble_words(words);  // store to global

      t.stop();

      locate_terms(w_tp, words);

      if (have_results && show_timing)  // globals
        disp_pct_timings();

      rw::named_timings.clear();
    }
  }
}

/**-----------------------------------------------------------------------------
 @brief    Clean & store the words from the string message 'post_msg' into the
 word@post map 'io_w_tp'.

           Take the input string 'post_msg', break it apart into individual
 words & clean them, then potentially store them each as a unique key into a map
 'io_w_tp'. The value for each key is a set-collection of paired
 thread_id+post_id ('thrd_id' + 'post_id')s. This effectively maps every single
 unique word to every IB post where it occurs at least once.

           Additionally, assemble all the cleaned words into another map (keyed
 by post ID) containing all the cleaned text of each post in the JSON files
 collection.

 @param    io_w_tp [IN/OUT] The word/thread+post map to store cleaned text into.
 @param    post_msg [IN] The text of the post's message.
 @param    thrd_id [IN] The thread's ID number.
 @param    post_id [IN] The post's ID number.
*/
void clean_store_words(Map_words_2_posts& io_w_tp,
                       const std::string& post_msg,
                       const uint32_t     thrd_id,
                       const uint32_t     post_id) {
  const auto words = scrub_n_extract(post_msg);

  for (const auto& word : words)
    io_w_tp[word].insert(make_pair(thrd_id, post_id));

  all_msgs[post_id] = rw::assemble_words(words);  // store to global
}

/**-----------------------------------------------------------------------------
 @brief    Load the word@post map 'io_w_tp', using each JSON file located inside
 the 'all_jsons' directory.

 @param    io_w_pt [IN/OUT] The word-counting map to load.
*/
void ld_map(Map_words_2_posts& io_w_tp) {
  //
  // lambda to parse a single JSON file into the word@post map
  auto parse_json_file{[&](auto& filepath) {
    const auto thread =
        rw::get_jsonfile_dat(filepath);  // each filepath == 1 imageboard thread

    const uint32_t thrd_id{thread["threadId"].asUInt()};
    const auto     thrd_subj = thread["subject"].asString();

    // store to global while here
    thrd_subjs[thrd_id] = rw::clean_str(thrd_subj);

    // OP's subject+message text
    clean_store_words(io_w_tp,
                      string{thrd_subj + ' ' + thread["message"].asString()},
                      thrd_id, thrd_id);

    // each reply post's subject+message text
    for (const auto& post : thread["posts"])
      clean_store_words(
          io_w_tp,
          string{post["subject"].asString() + ' ' + post["message"].asString()},
          thrd_id, post["postId"].asUInt());
  }};

  // first, read in all JSON file paths from the directory
  vector<file_path> paths{};
  for (const auto& jsonfile : dir_iter{jsons_dir})  // jsons_dir{"./all_jsons/"}
    paths.push_back(jsonfile);

  // next, run file processing calling the lambda above
  for_each(
#if __cpp_lib_parallel_algorithm >= 201603
      par_unseq,
#endif
      cbegin(paths), cend(paths), parse_json_file);
}

/**-----------------------------------------------------------------------------
 @brief    Stream the vector 'vw_vtp' out to the full word/post/count file.

 @param    vw_vtp [IN] The vector to stream to file.
*/
void wrt_counts_file(const Vec_words_2_posts& vw_vtp) {
  auto ofs = rw::write_file_ready("word_post_counts.txt");

  for (const auto& [term, thrd_post] : vw_vtp)
    ofs << left << setw(7) << thrd_post.size() << term << '\n';
}

/**-----------------------------------------------------------------------------
 @brief    Copy word@posts map to vector & write that out to a descending-sorted
 words+counts file.

 @param    w_tp [IN] The map to write out to sorted file.
*/
void sort_wrt_counts(const Map_words_2_posts& w_tp) {
  // disable general timing
  bool temp     = rw::do_timing;
  rw::do_timing = false;

  Vec_words_2_posts vw_vtp_counts{};
  cp_map_2_vec(w_tp, vw_vtp_counts);
  sort_counts(vw_vtp_counts, true);
  wrt_counts_file(vw_vtp_counts);

  // restore general timing state
  rw::do_timing = temp;
}

/**-----------------------------------------------------------------------------
 @brief    Write the 'all_posts' map containing all the cleaned text of all the
 posts out to file.
*/
void wrt_all_posts() {
  auto ofs = rw::write_file_ready("all_posts.txt");

  for (const auto& [post_id, post_msg] : all_msgs)
    ofs << post_id << '\n' << post_msg << "\n\n";
}

/**-----------------------------------------------------------------------------
 @brief    The program entry point.

 @param    argc [IN] The CLI arguments count.
 @param    argv [IN] The CLI arguments C-string array.
 @returns  The program's exit status.
*/
int main(int argc, char** argv) {
  rw::parse_args(argc, argv);

  auto bgn_ldwrt = rw::Clock::now();

  // load the full word/thread_id+post_id (word@post) map from JSON files
  Map_words_2_posts w_tp{};
  ld_map(w_tp);

  // DESIGN: create a task-system, and then just queue the two file writes?
  if (write_files) {
    wrt_all_posts();        // write all post's cleaned plaintext out to file
    sort_wrt_counts(w_tp);  // write sorted word counts out to file
  }

  cout << "time to process local /" + board_nm + "/ JSON data:  "
       << duration_cast<Ms>(rw::Clock::now() - bgn_ldwrt).count() << " ms\n";

  do_search(w_tp);  // search-loop the word@post map
}

// Copyright (2021)
// License (MIT) https://opensource.org/licenses/MIT
