// waifusearch, the IB searcher
// ============================
// -This software is designed to search an imageboard's text.

// -Filename: rw_cli_args.hpp
// -Command-line parsing & help support functions.

#pragma once

#include <cstdlib>
#include <string>

#include "CLI11.hpp"
#include "user_flags.hpp"

namespace rw {

// This is the single, and overall CLI App object for parsing program arguments
CLI::App cli_app{"waifusearch, the IB searcher\n (q to quit)\n"};

// Simple flag to track if the help flag was passed in
bool have_help{false};

/**-----------------------------------------------------------------------------
 @brief    Configure for, and parse, any command-line arguments.

 @param    argc [IN] The CLI arguments count.
 @param    argv [IN] The CLI arguments C-string array.
 @returns  Success code 0, or a parse error code.
*/
int handle_args(int argc, char** argv) {
  cli_app.add_option("-y, --hyperlinks", make_hyperlinks,
                     "Make hyperlinks (default:true, false==crosslinks)");
  cli_app.add_option("-t, --timing", show_timing,
                     "Show search timings (default:false)");
  cli_app.add_option("-e, --terms", show_terms,
                     "Show search term counts (default:false)");
  cli_app.add_option("-w, --write", write_files,
                     "Write files out on startup (default:true)");

  // https://github.com/CLIUtils/CLI11/blob/master/examples/modhelp.cpp
  // Disable the default help handler...
  cli_app.set_help_flag();
  // and add a custom flag for activating help
  auto help = cli_app.add_flag("-h, --help", "Display this help and quit");

  try {
    cli_app.parse(argc, argv);

    if (*help) {                 // Special block just for the custom help flag
      have_help = true;          // used by parse_args() below
      throw CLI::CallForHelp();  // caught immediately below, returns a '0' exit
    }
  } catch (const CLI::ParseError& e) {
    return cli_app.exit(e);
  }

  return 0;
}

/**-----------------------------------------------------------------------------
 @brief    Wrapper to unify correct program exit on either arg errors or the
 help flag.
*/
void parse_args(int argc, char** argv) {
  if (auto bad_arg = handle_args(argc, argv)) {  // non-zero ret == error state
    std::exit(bad_arg);
  } else if (have_help) {  // simple hack to prevent program execution on --help
    std::exit(0);
  }

  // DESIGN: This would probably be a good place to emplace CLI option pointers
  // into a general-purpose std::vector<Option*> (if CLI doesn't already provide
  // something like this) and pass it back to caller. Possibly a better approach
  // than keeping 'global' vars scattered all over inside this namespace as now.
  // -Using a vector instead would allow us to efficiently manage an unlimited
  // number of arguments passed into our program once they are parsed, It would
  // be a simple handle to them all, easily usable from anywhere in our code.

  // ...
}

/**-----------------------------------------------------------------------------
 @brief    Display the help before exit.
*/
void help_exit() {
  //
  cli_app.exit(CLI::CallForHelp());
  std::exit(0);
}

}  // namespace rw

// Copyright (2021)
// License (MIT) https://opensource.org/licenses/MIT
