// waifusearch, the IB searcher
// ============================
// -This software is designed to search an imageboard's text.

// -Filename: rw_text_utils.hpp
// -Various text-processing & other support functions.

#pragma once

#include <algorithm>
#include <cmath>
#include <filesystem>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <iterator>
#include <regex>
#include <sstream>
#include <string>
#include <utility>

#if __has_include(<execution>)  // RPi will only run single-threaded for now
#include <execution>
using std::execution::par_unseq;
#endif

#if __has_include(<jsoncpp/json/json.h>)  // RPi's version
#include <jsoncpp/json/json.h>
#else
#include <json/json.h>
#endif

// alias'
namespace fs    = std::filesystem;
using os_it_str = std::ostream_iterator<std::string>;

using std::begin;
using std::cend;
using std::copy;
using std::end;
using std::ifstream;
using std::make_pair;
using std::ofstream;
using std::regex;
using std::regex_replace;
using std::string;
using std::stringstream;
using std::to_string;

namespace rw {

/**-----------------------------------------------------------------------------
 @brief    Split 'in' string into a pair, dividing on the final 'delim' char.

 @param    in [IN] The input string to split.
 @param    delim [IN] The delimiter char to divide upon.
 @returns  The now-split string pair.
*/
std::pair<std::string, std::string> split_str(const std::string& in,
                                              const char         delim) {
  const auto str_end{string::npos};  // note: this works in both directions

  if (const auto split_pos{in.rfind(delim)};  // Search reverse for final delim
      split_pos != str_end) {                 // Delimeter found
    const string front{in.substr(0, split_pos)};
    const string back{in.substr(split_pos + 1, str_end)};
    return make_pair(front, back);
  } else
    return make_pair(in, "");  // Delimeter not found
}

/**-----------------------------------------------------------------------------
 @brief    split_str() wrapper that defaults using a '.' delimiter char.

 @param    in [IN] The input string to split.
 @returns  The now-split string pair.
*/
std::pair<std::string, std::string> split_str_dot(const std::string& in) {
  return split_str(in, '.');
}

/**-----------------------------------------------------------------------------
 @brief    split_str() wrapper that defaults using a '/' delimiter char.

 @param    in [IN] The input string to split.
 @returns  The now-split string pair.
*/
std::pair<std::string, std::string> split_str_slash(const std::string& in) {
  return split_str(in, '/');
}

/**-----------------------------------------------------------------------------
 @brief    Convert all alphabetic characters of 'in_out' string in-place, to
 lowercase.

 @param    in_out [IN/OUT] The string to convert.
*/
void to_lower(std::string& in_out) {
  std::transform(
      begin(in_out), end(in_out), begin(in_out),
      [](unsigned char c) -> unsigned char { return std::tolower(c); });
}

/**-----------------------------------------------------------------------------
 @brief    Replace all bad characters of 'in_out' string in-place, with a
 specified good char instead.

 @param    in_out [IN/OUT] The string to process.
 @param    bad [IN] The character to replace.
 @param    good [IN] The replacement character.
*/
void repl_bad_char(std::string& in_out, const char bad, const char good) {
  std::transform(
      begin(in_out), end(in_out), begin(in_out),
      [=](unsigned char c) -> unsigned char { return c == bad ? good : c; });
}

/**-----------------------------------------------------------------------------
 @brief    Remove specific substring[s] 'sub' from 'in_out' string, in-place.

 @param    in_out [IN/OUT] The string to process.
 @param    sub [IN] The substring[s] to remove.
*/
void rm_substrings(std::string& in_out, const std::string& sub) {
  const auto str_end{string::npos};
  const auto sub_len{sub.length()};

  for (auto i = in_out.find(sub); i != str_end; i = in_out.find(sub)) {
    in_out.erase(i, sub_len);
  }
}

/**-----------------------------------------------------------------------------
 @brief    Consolidate 'in_out' string in-place, paring contiguous chars 'c'
 down to just one 'c' per contiguous sequence.

 @param    in_out [IN/OUT] The string to pare down.
 @param    c [IN] The char to consolidate upon.
*/
void rm_dupes(std::string& in_out, const char c) {
  char prev{'\0'};

  for (auto it{begin(in_out)}; it != cend(in_out); ++it) {
    if ((prev == *it) && (c == *it)) {  // Have a duplicate char 'c'
      in_out.erase(it);                 // remove it,
      --it;                             // now back up to adjust for next check
    } else
      prev = *it;
  }
}

/**-----------------------------------------------------------------------------
 @brief    Edit 'in_out' string in-place, removing each single leading char 'c',
 if present.

 @param    in_out [IN/OUT] The input string to trim.
 @param    c [IN] The target char to trim from string.
*/
void trim_l(std::string& in_out, const char c) {
  auto it{begin(in_out)};
  if (*it == c)
    in_out.erase(it);
}

/**-----------------------------------------------------------------------------
@brief    Edit 'in_out' string in-place, removing each single trailing char 'c',
if present.

 @param    in_out [IN/OUT] The input string to trim.
 @param    c [IN] The target char to trim from string.
*/
void trim_r(std::string& in_out, const char c) {
  auto it{end(in_out) - 1};
  if (*it == c)
    in_out.erase(it);
}

/**-----------------------------------------------------------------------------
 @brief    Edit 'in_out' string in-place, removing each single leading and
 trailing char 'c', if present.

 @param    in_out [IN/OUT] The input string to trim.
 @param    c [IN] The target char to trim from string.
*/
void trim_lr(std::string& in_out, const char c) {
  trim_l(in_out, c);
  trim_r(in_out, c);
}

/**-----------------------------------------------------------------------------
 @brief    Edit 'in_out' string in-place, first consolidating (remove dupes),
 then removing all leading and trailing chars 'c', if present.

 @param    in_out [IN/OUT] The input string to trim.
 @param    c [IN] The target char to trim from string.
*/
void squash_n_trim(std::string& in_out, const char c) {
  rm_dupes(in_out, c);
  trim_lr(in_out, c);
}

/**-----------------------------------------------------------------------------
 @brief    Take the vector of strings 'words' and assemble each one,
 space-separated, into a new single string.

 @param    words [IN] The vector of strings to assemble.
 @returns  The assembled string.
*/
std::string assemble_words(const std::vector<std::string>& words) {
  stringstream ss{};

  copy(
#if __cpp_lib_parallel_algorithm >= 201603
      par_unseq,
#endif
      cbegin(words), cend(words), os_it_str(ss, " "));

  auto res = ss.str();
  rw::trim_r(res, ' ');
  return res;
}

//////////////
// REWORK: No copies, if possible
//

// DESIGN: convert these to C++20 std::format when available

/**-----------------------------------------------------------------------------
 @brief    Return a new string, padding the 'str' string out to 'n' length if
 necessary, using space chars. Truncate it to 'n' length otherwise.

 @param    str [IN] The input source string.
 @param    n [IN] The target string length.
 @param    r_align [IN] A flag determining alignment mode for string padding.
 @returns  The new, padded string.
*/
std::string make_pad_str(const std::string& str,
                         const uint32_t     n       = 25,
                         const bool         r_align = false) {
  string res{};

  if (str.size() > n)
    res = str.substr(0, n);
  else {
    stringstream ss{};

    if (r_align)
      ss << std::setfill(' ') << std::setw(n) << str;
    else
      ss << std::setfill(' ') << std::setw(n) << std::left << str;

    res = ss.str();
  }

  return res;
}

//////////////
// COPY-ONLY
// -Standard library regex-based functions
//

/**-----------------------------------------------------------------------------
 @brief    Return a copy of 'in', replacing each bad substring with a single
 good character.

 @param    in [IN] The input source string.
 @param    bad [IN] The bad search pattern string.
 @param    good [IN] The good replacement char.
 @returns  The now-cleaned string copy.
*/
std::string repl_w_char(const std::string& in,
                        const std::string& bad,
                        const char         good) {
  const regex  search{bad};
  const string replace{good};
  return regex_replace(in, search, replace);
}

/**-----------------------------------------------------------------------------
 @brief    Return a copy of 'in', replacing each bad substring 'bad' with an
 alternate string 'good'.

 @param    in [IN] The input source string.
 @param    bad [IN] The bad search pattern string.
 @param    good [IN] The good replacement string.
 @returns  The now-cleaned string copy.
*/
std::string repl_w_str(const std::string& in,
                       const std::string& bad,
                       const std::string& good) {
  const regex search{bad};
  return regex_replace(in, search, good);

  // DESIGN: Possibly could be reworked with a rm_bad_string()-like algorithm
  // instead. (no-copy)
}

/**-----------------------------------------------------------------------------
 @brief    Return a copy of 's', cleaning various bad JSON text elements it may
 contain (currently <&apos;> & <&quot;>) .

 @param    s [IN] The string to clean.
 @returns  The cleaned string.
*/
std::string clean_str(const std::string& s) {
  std::string res{s};
  // handle individual JSON string-to-char replacements
  res = rw::repl_w_char(res, "&apos;", '\'');
  res = rw::repl_w_char(res, "&quot;", '"');

  return res;
}

//////////////
// OTHER:
//

/**-----------------------------------------------------------------------------
 @brief    Calculate a percentage from the 'num_part' & 'den_whole' components,
 returning a rounded, formatted string version of the result.

 @param    num_part [IN] The numerator (part) of the percentage calc.
 @param    den_whole [IN] The denominator (whole) of the percentage calc.
 @param    pad_len [IN] A number specifying the output string length.
 @returns  The rounded, formatted string.
*/
std::string make_pct_str(const double   num_part,
                         const double   den_whole,
                         const uint32_t pad_len = 2) {
  if (den_whole != 0.0) {
    stringstream ss{};
    ss << std::setfill(' ') << std::setw(pad_len)
       << std::round((num_part / den_whole) * 100);
    return ss.str();
  } else
    return "**";
}

/**-----------------------------------------------------------------------------
 @brief    Attempt to open a valid output stream 'filepath' to disk.

 @param    filepath [IN] The file pathname of the target file.
 @returns  The validated output file stream.
*/
std::ofstream write_file_ready(const fs::path& filepath) {
  ofstream ofs{filepath};
  if (fs::exists(filepath) && fs::is_regular_file(filepath)) {
    return ofs;
  } else {
    std::cerr << "\n\nfile error, write_file_ready()\n\n";
    std::exit(EXIT_FAILURE);
  }
}

/**-----------------------------------------------------------------------------
 @brief    Take a raw JSON source and parse it into a Json DB.

 @param    json_src [IN] The raw input JSON source.
 @returns  The Json data after parsing.
*/
template <typename T>
Json::Value parse_json(T& json_src) {
  Json::Reader reader{};
  Json::Value  json_dat{};
  reader.parse(json_src, json_dat);
  return json_dat;
}

/**-----------------------------------------------------------------------------
 @brief    Validate 'filepath' then return the file's Json parse.

 @param    filepath [IN] The input filesystem path.
 @returns  The Json data after parsing.
*/
Json::Value get_jsonfile_dat(const fs::path& filepath) {
  if (fs::exists(filepath) && fs::is_regular_file(filepath)) {
    ifstream jsonfile{filepath.string()};
    return parse_json(jsonfile);
  } else
    return Json::Value{};
}

}  // namespace rw

// Copyright (2021)
// License (MIT) https://opensource.org/licenses/MIT
